@@ -499,8 +499,7 @@ nm_dhcp_manager_handle_timeout (gpointer user_data)
 {
 	NMDHCPDevice *device = (NMDHCPDevice *) user_data;
 
-	nm_info ("Device '%s' DHCP transaction took too long (>%ds), stopping it.",
-			 device->iface, NM_DHCP_TIMEOUT);
+	nm_info ("(%s): DHCP transaction took too long, stopping it.", device->iface);
 
 	nm_dhcp_manager_cancel_transaction (device->manager, device->iface);
 
@@ -577,8 +576,10 @@ static void dhcp_watch_cb (GPid pid, gint status, gpointer user_data)
 gboolean
 nm_dhcp_manager_begin_transaction (NMDHCPManager *manager,
                                    const char *iface,
+                                   const char *uuid,
                                    NMSettingIP4Config *s_ip4,
-                                   guint32 timeout)
+                                   guint32 timeout,
+                                   guint8 *dhcp_anycast_addr)
 {
 	NMDHCPManagerPrivate *priv;
 	NMDHCPDevice *device;
@@ -597,15 +598,16 @@ nm_dhcp_manager_begin_transaction (NMDHCPManager *manager,
 		nm_dhcp_manager_cancel_transaction_real (device);
 	}
 
-	nm_info ("Activation (%s) Beginning DHCP transaction.", iface);
-
-	device->pid = nm_dhcp_client_start (device, s_ip4);
-	if (device->pid == 0)
-		return FALSE;
-
 	if (timeout == 0)
 		timeout = NM_DHCP_TIMEOUT;
 
+	nm_info ("Activation (%s) Beginning DHCP transaction (timeout in %d seconds)",
+	         iface, timeout);
+
+	device->pid = nm_dhcp_client_start (device, uuid, s_ip4, dhcp_anycast_addr);
+	if (device->pid == 0)
+		return FALSE;
+
 	/* Set up a timeout on the transaction to kill it after the timeout */
 	device->timeout_id = g_timeout_add_seconds (timeout,
 	                                            nm_dhcp_manager_handle_timeout,
@@ -681,9 +683,8 @@ nm_dhcp_manager_cancel_transaction_real (NMDHCPDevice *device)
 		device->pid_file = NULL;
 	}
 
-	/* Clean up the leasefile if it got left around */
+	/* Free leasefile (but don't delete) */
 	if (device->lease_file) {
-		remove (device->lease_file);
 		g_free (device->lease_file);
 		device->lease_file = NULL;
 	}
@@ -771,6 +772,45 @@ out:
 	g_strfreev (searches);
 }
 
+static void
+process_domain_search (NMIP4Config *ip4_config, const char *str)
+{
+	char **searches, **s;
+	char *unescaped, *p;
+	int i;
+
+	g_return_if_fail (str != NULL);
+	g_return_if_fail (ip4_config != NULL);
+
+	p = unescaped = g_strdup (str);
+	do {
+		p = strstr (p, "\\032");
+		if (!p)
+			break;
+
+		/* Clear the escaped space with real spaces */
+		for (i = 0; i < 4; i++)
+			*p++ = ' ';
+	} while (*p++);
+
+	if (strchr (unescaped, '\\')) {
+		nm_info ("  invalid domain search: '%s'", unescaped);
+		goto out;
+	}
+
+	searches = g_strsplit (unescaped, " ", 0);
+	for (s = searches; *s; s++) {
+		if (strlen (*s)) {
+			nm_info ("  domain search '%s'", *s);
+			nm_ip4_config_add_search (ip4_config, *s);
+		}
+	}
+	g_strfreev (searches);
+
+out:
+	g_free (unescaped);
+}
+
 /* Given a table of DHCP options from the client, convert into an IP4Config */
 NMIP4Config *
 nm_dhcp_manager_options_to_ip4_config (const char *iface, GHashTable *options)
@@ -882,16 +922,8 @@ nm_dhcp_manager_options_to_ip4_config (const char *iface, GHashTable *options)
 	}
 
 	str = g_hash_table_lookup (options, "new_domain_search");
-	if (str) {
-		char **searches = g_strsplit (str, " ", 0);
-		char **s;
-
-		for (s = searches; *s; s++) {
-			nm_info ("  domain search '%s'", *s);
-			nm_ip4_config_add_search (ip4_config, *s);
-		}
-		g_strfreev (searches);
-	}
+	if (str)
+		process_domain_search (ip4_config, str);
 
 	str = g_hash_table_lookup (options, "new_netbios_name_servers");
 	if (str) {
@@ -917,7 +949,7 @@ nm_dhcp_manager_options_to_ip4_config (const char *iface, GHashTable *options)
 		if ((errno == EINVAL) || (errno == ERANGE))
 			goto error;
 
-		if (int_mtu)
+		if (int_mtu > 576)
 			nm_ip4_config_set_mtu (ip4_config, int_mtu);
 	}
 
